"""
Code generator for typescript object representing the entity instances
"""

import codecs
from pathlib import Path
from typing import Dict

from dmt.common.package import Package
from dmt.common.blueprint import Blueprint
from dmt.common.blueprint_attribute import BlueprintAttribute
from dmt.package_generator import PackageGenerator
from dmt.template_generator import TemplateBasedGenerator
from jinja2.environment import Template


class EntityGenerator(TemplateBasedGenerator):
    """Generates entities"""

    types = {"number": "number", "double": "number", "string": "string", "char": "string",
            "integer": "number", "short": "number", "boolean": "boolean"}

    default_values = {"number": "0.0", "string": "\"\"", "int": "0", "boolean": "false"}

    def generate(
        self,
        package_generator: PackageGenerator,
        template: Template,
        outputfile: Path,
        config: Dict,
    ):
        """Generate typescript classes with properties"""
        if self.__skip(template,config):
            return
        outputdir = outputfile.parents[0]
        root_package = package_generator.root_package
        self.__generate_package(root_package,template, outputdir, config)

    def __skip(self, template: Template, config: Dict):
        is_poto = template.name == "poto.ts.jinja"
        if config.get("minimal", False):
            # This is the minimal stage
            return is_poto

        return not is_poto

    def __generate_package(self, package: Package, template, pkg_dir,config: Dict):
        for blueprint in package.blueprints:
            self.__generate_entity(blueprint,template,pkg_dir)

        for package in package.packages:
            name = package.name
            sub_dir = pkg_dir / name
            self.__generate_package(package,template, sub_dir, config)

    def __generate_entity(self, blueprint: Blueprint,template: Template, outputdir: Path):
        outputdir.mkdir(exist_ok=True)
        package=blueprint.get_parent()
        model = self.__create_model(blueprint, package.name)
        filename = model["file_basename"].lower() + ".ts"
        outputfile = outputdir / filename
        with codecs.open(outputfile, "w", "utf-8") as file:
            file.write(template.render(model))

    def __create_model(self, blueprint: Blueprint, package_name):
        model = {}
        name = blueprint.name
        model["package"] = package_name
        model["package_class"] = (
            self.__first_to_upper(package_name) + "PackageDescription"
        )

        model["imports"] = self.__find_imports(blueprint)
        model["extends"] = self.__find_extensions(blueprint)
        model["name"] = name
        model["file_basename"] = name.lower()
        model["description"] = self.__reshape_multiline_string(
            blueprint.description
        )
        model["type"] = self.__first_to_upper(name)

        fields = []
        # for attribute in blueprint.all_attributes.values():
        for attribute in blueprint.attributes:
            field = self.__create_field(attribute)
            if field:
                fields.append(field)

        model["fields"] = fields

        return model

    def __find_extensions(self, blueprint: Blueprint) -> str:
        extensions = []
        for ext in blueprint.extensions:
            if ext.name == "NamedEntity":
                continue
            extensions.append(self.__first_to_upper(ext.name))

        if len(extensions) == 0:
            return ""
        return "extends " +  ", ".join(extensions)

    def __find_imports(self, blueprint: Blueprint) -> str:
        imports = []
        for ext in blueprint.extensions:
            if ext.name == "NamedEntity":
                continue
            ext_name =self.__first_to_upper(ext.name)
            fname = ext.name.lower()
            imports.append(f"import {{{ext_name}}} from './{fname}'")

        return imports

    def __reshape_multiline_string(self, value: str) -> str:
        if not value:
            return ""
        return value.replace("'", '"').encode("unicode_escape").decode("utf-8")

    def __first_to_upper(self, string):
        # Make sure the first letter is uppercase
        return string[:1].upper() + string[1:]

    def __create_field(self, attribute: BlueprintAttribute):
        field = {}
        field["name"] = attribute.name
        is_array = attribute.is_many
        field["isArray"] = is_array

        a_type: str = attribute.get("attributeType")
        if a_type not in self.types:
            # Blueprint, FIXME
            return None
        ftype = self.__map(attribute.type, self.types)

        if is_array:
            field["type"] = "BaseList<" + ftype + ">"
            field["init"] = "new BaseList()"
        else:
            field["type"] = ftype
            field["init"] = self.__find_default_value(attribute)

        field["createMethod"] = self.__first_to_upper(ftype)
        field["description"] = self.__reshape_multiline_string(attribute.get("description", ""))
        return field

    def __map(self, key, values):
        converted = values[key]
        if not converted:
            raise Exception("Unkown type " + key)
        return converted
    
    def __map_type(self,ptype):
        return self.__map(ptype, self.types)

    def __find_default_value(self, attribute: BlueprintAttribute):
        """Returns the default value literal"""
        a_type: str = attribute.get("attributeType")
        etype = self.__map_type(a_type)
        default_value = attribute.get("default")
        if default_value is not None:
            return self.__convert_default(attribute,default_value)
        default_value = self.__map(etype, self.default_values)
        return default_value


    def __convert_default(self, attribute: BlueprintAttribute, default_value):
        # converts json value to Python value
        if isinstance(default_value,str):
            if default_value == '' or default_value == '""':
                return '""'
            elif attribute.type == 'integer':
                return int(default_value)
            elif attribute.type == 'number':
                return float(default_value)
            elif attribute.type == 'boolean':
                bval = bool(default_value)
                if bval:
                    return "true"
                return "false"
            else:
                return "'" + default_value + "'"
        return default_value
